// PxM.h
// (c) 2001-2004 by Jeremy Stanley
// This is free sofwtare licensed under the GNU GPL; see license.txt for details.

#ifndef PxM_H
#define PxM_H

#include <string.h>
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>

// PxM image class

class PxM
{
    // image data and characteristics
    unsigned char *pixels;
    int width;
    int height;
    int iscolor;
	int bytes;
    int bytes_per_pixel;
    int pitch;

    // helpful functions
    char *getToken(char *, int, FILE *);

public:
    PxM();
    PxM(const PxM &other);
    PxM(int width, int height, bool color = true);
    ~PxM();

	void init(int width, int height, bool color = true);
	void reset();

	void pixel(int u, int v, int col[3]);
    void bilinearpix(double u, double v, int col[3]);
    void nearestpix(double u, double v, int col[3]);

	void setpixel(int u, int v, const int col[3]);

    int getWidth() { return width; }
    int getHeight() { return height; }
    int getPitch() { return pitch; }
    unsigned char *getPixels() { return pixels; }
    int isColor() { return iscolor; }

    int load(const char *filename);

	int load_PxM(const char *filename);
	int load_PxM(FILE *file);
    int save_PxM(const char *filename);

#ifdef USE_JPEG_LIB
	int load_jpeg(const char *filename);
	int save_jpeg(const char *filename, int quality = 80);
#endif

    PxM &operator= (const PxM &other);
    int subtract(PxM &other);
    int negate();
	int scale(int width, int height);
	int convertbw();

#ifdef USE_GRAPHICS_H
	void draw(int x, int y);
    void draw(int x0, int y0, int x1, int y1);
    void plot_rotate(int x, int y, double angle);
    void plot_center(int x0, int y0);
    int get(int x0, int y0, int w, int h, int iscolor = 1); // captures a section of the window
#endif

};


#endif // PxM_H
