// coerce.cpp - The Computer Coercion Tool
// (c) 2004 by Jeremy Stanley
// This is free software licensed under the GNU GPL.  See license.txt for details.

#pragma warning ( disable : 4786 )

#ifdef WIN32
#include "SDL_syswm.h"
#include "resource.h"
#endif

#include <time.h>
#include <stdlib.h>
#include <math.h>
#include "ps-sdl_gl_app.h"
#include "PxM.h"

using namespace std;

static const double PI = 3.141592653589793238462643383279502884197169399375105820; // close enough, I guess...

PxM g_picBackground;

class banger : public psSdlGlApp
{
	GLuint m_hammer_handle, m_hammer_topper;
	GLuint m_background, m_wood;
	GLfloat m_background_width, m_background_height;
	GLdouble m_angle;
	GLdouble m_x, m_y;
	GLdouble m_mousex, m_mousey;
	GLdouble m_hitx, m_hity;
	GLdouble m_bx, m_by, m_bz;
	GLfloat  m_clr, m_clg, m_clb;
	GLfloat  m_rr, m_rg, m_rb;
	GLdouble m_swing_time;

	SDL_AudioSpec *m_wav_spec;
	Uint8 *m_wav_buf;

	enum state {
		idle,
		downswing,
		upswing
	};
	state m_state;

	template <class F>
	static F frand(F a, F b)
	{
		return ((F)rand() / RAND_MAX) * (b - a) + a;
	}

	void cylinder(GLdouble radius, GLdouble length, int segment_count)
	{
		GLdouble theta = 0;
		GLdouble dtheta = (2 * PI) / (segment_count - 1);
		int i;

		glBegin(GL_QUAD_STRIP);
		for(i = 0; i < segment_count; i++)
		{
			GLdouble X = radius * cos(theta);
			GLdouble Y = radius * sin(theta);
			theta += dtheta;

			glNormal3d(X, Y, 0);

			GLfloat t = 1.0f - (theta / (2 * PI));

			glTexCoord2f(1.0, t);
			glVertex3d(X, Y, length);

			glTexCoord2f(0.0, t);
			glVertex3d(X, Y, 0);
		}
		glEnd();

		// cap 1
		glBegin(GL_TRIANGLE_FAN);
		glNormal3d(0, 0, -1);
		glVertex3d(0, 0, 0);
		for(i = 0; i < segment_count; i++)
		{
			GLdouble X = radius * cos(theta);
			GLdouble Y = radius * sin(theta);
			theta -= dtheta;

			glVertex3d(X, Y, 0);
		}
		glEnd();

		// cap 2
		glBegin(GL_TRIANGLE_FAN);
		glNormal3d(0, 0, 1);
		glVertex3d(0, 0, length);
		for(i = 0; i < segment_count; i++)
		{
			GLdouble X = radius * cos(theta);
			GLdouble Y = radius * sin(theta);
			theta += dtheta;

			glVertex3d(X, Y, length);

		}
		glEnd();
		

	}
	
	void build_lists()
	{
		m_hammer_handle = glGenLists(1);
		glNewList(m_hammer_handle, GL_COMPILE);
		glRotated(90.0, 1.0, 0.0, 0.0);
		glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
			cylinder(0.35, 10.0, 30);
		glEndList();

		m_hammer_topper = glGenLists(1);
		glNewList(m_hammer_topper, GL_COMPILE);
			glColor4f(0.33f, 0.33f, 0.33f, 1.0f);
			glRotated(90.0, 1.0, 0.0, 0.0);
			glTranslated(0.0, 8.25, -2.0);
			cylinder(1.50, 4, 50);
		glEndList();

		m_background = CreateTexture(g_picBackground);
		m_wood       = CreateTexture("wood.ppm");
	}

	struct audio_callback_data
	{
		Uint8 *buf;
		Uint32 size;
		Uint32 pos;
	} m_audio_callback_data;

	static void callback(void *udata, Uint8 *stream, int len)
	{
		audio_callback_data *dat = (audio_callback_data *)udata;

		Uint32 rem = dat->size - dat->pos;
		if (rem == 0)
			return;
		if (len > rem)
			len = rem;
		SDL_MixAudio(stream, dat->buf + dat->pos, len, SDL_MIX_MAXVOLUME);
		dat->pos += len;
	}

	void init_noise()
	{
		m_wav_spec = (SDL_AudioSpec *)malloc(sizeof(SDL_AudioSpec));
		SDL_LoadWAV("coerce.wav", m_wav_spec, &m_wav_buf, &m_audio_callback_data.size);
		if ( m_wav_spec->freq <= 16000 )
			m_wav_spec->samples = 256;
		else
			m_wav_spec->samples = 512;
		m_wav_spec->callback = callback;
		m_wav_spec->userdata = &m_audio_callback_data;
		SDL_OpenAudio(m_wav_spec, NULL);

		m_audio_callback_data.buf = m_wav_buf;
		m_audio_callback_data.pos = 0;
	}

	void play_noise()
	{
		m_audio_callback_data.pos = 0;
		SDL_PauseAudio(0);
	}

	void uninit_noise()
	{
		SDL_FreeWAV(m_wav_buf);
		free(m_wav_spec);
	}

	void draw_background()
	{
		glBindTexture(GL_TEXTURE_2D, m_background);

		if ( m_state == upswing )
		{
			glRotated((m_hitx * m_hitx + m_hity * m_hity) * m_swing_time, -m_hity, m_hitx, 0);
			glTranslated(m_bx, m_by, m_bz);
			glColor3f(m_clr, m_clg, m_clb);
		}
		else
		{
			glColor3f(1.0f, 1.0f, 1.0f);
		}

		glBegin(GL_QUADS);
		glTexCoord2f(0, 1 - m_background_height);						glVertex3d( -8, -6,  0);
		glTexCoord2f(m_background_width, 1 - m_background_height);		glVertex3d(  8, -6,  0);
		glTexCoord2f(m_background_width, 1);							glVertex3d(  8,  6,  0);
		glTexCoord2f(0, 1);												glVertex3d( -8,  6,  0);
		glEnd();

		glBindTexture(GL_TEXTURE_2D, 0);
	}

public:
	banger(GLfloat tw, GLfloat th) :
		m_background_width(tw),
		m_background_height(th)
	{
	}

	~banger()
	{
		uninit_noise();
	}

	void Initialize()
	{
		srand( (unsigned)time(0) );

		// initialize some GL stuff
		glClearColor(0.0, 0.0, 0.0, 0.0);
		glEnable(GL_TEXTURE_2D);

		// set up material color tracking
		glColorMaterial(GL_FRONT, GL_AMBIENT_AND_DIFFUSE);
		glEnable(GL_COLOR_MATERIAL);
		 
		// alpha blending
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		glEnable(GL_BLEND);

		// lighting
		GLfloat lp[4];

		lp[0] = 0.25f; lp[1] = 0.25f; lp[2] = 0.25f; lp[3] = 1.00f;
		glLightfv(GL_LIGHT0, GL_AMBIENT, lp);

		lp[0] = 0.75f; lp[1] = 0.75f; lp[2] = 0.75f; lp[3] = 1.00f;
		glLightfv(GL_LIGHT0, GL_DIFFUSE, lp);

		lp[0] = 1.00f; lp[1] = 0.00f; lp[2] = 0.00f; lp[3] = 1.00f;
		glLightfv(GL_LIGHT0, GL_SPECULAR, lp);

		glShadeModel(GL_SMOOTH);
		glEnable(GL_LIGHTING);
		glEnable(GL_LIGHT0);

		// depth test
		glDepthFunc(GL_LEQUAL);
		glEnable(GL_DEPTH_TEST);

		// normalize 
		glEnable(GL_NORMALIZE);

		// enable back face culling
		glCullFace(GL_BACK);
		glEnable(GL_CULL_FACE);

		// set up texture env stuff
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
		glEnable(GL_TEXTURE_2D);

		// let game initialize logical width and height
		Reshape(m_Width, m_Height);

		// setup viewport
		// (16 logical units wide by 12 logical units tall,
		// with the vanishing point in the center of the screen at (0,0)
		glMatrixMode(GL_PROJECTION);
		
		glFrustum(-8.0, 8.0, -6.0, 6.0, 1.0, 10.0);
		glTranslated(0, 0, -2);
		glScaled(2.0, 2.0, 0.1);

		GLfloat lightpos[4] = { 0, 0, 8, 1 };
		glLightfv(GL_LIGHT0, GL_POSITION, lightpos);


		// build lists
		build_lists();

		// init state
		m_angle = 120.0;
		m_x = 0.0; m_mousex = 0.0;
		m_y = 0.0; m_mousey = 0.0;
		m_bx = 0;
		m_by = 0;
		m_bz = 0;
		m_clr = 1.0f;
		m_clg = 1.0f;
		m_clb = 1.0f;
		m_rr = 1.0f;
		m_rg = 1.0f;
		m_rb = 1.0f;
		m_swing_time = 0.0;
		m_state = idle;

		// init audio
		init_noise();

		SDL_ShowCursor(0);
	}

	void UpdateBusyState(double frametime)
	{
		if ( m_state == downswing )
		{
			m_swing_time += frametime;
			m_angle = 120.0 + m_swing_time * 60.0 / 0.10;

			if ( m_angle > 180.0 )
			{
				m_hitx = m_mousex;
				m_hity = m_mousey;
				
				play_noise();

				m_state = upswing;
				m_angle = 180.0;
				m_swing_time = 0.25;

				m_rr = 1.0f + frand(0.0f, 2.0f);
				m_rg = 1.0f + frand(0.0f, 2.0f);
				m_rb = 1.0f + frand(0.0f, 2.0f);
			}
		}
		else if ( m_state == upswing )
		{
			m_bz = -6.0 * m_swing_time;
			m_bx = frand(-1.5, 1.5) * m_swing_time;
			m_by = frand(-1.5, 1.5) * m_swing_time;
			m_clr = 1.0f - m_swing_time * m_rr;
			m_clg = 1.0f - m_swing_time * m_rg;
			m_clb = 1.0f - m_swing_time * m_rb;


			m_swing_time -= frametime;
			m_angle = 180.0 - ((0.25 - m_swing_time) * 60.0) / 0.25;
			if ( m_swing_time < 0.0 )
			{
				m_angle = 120.0;
				m_swing_time = 0.0;
				m_state = idle;
				EndBusyState();
			}
		}
	}

	void Draw()
	{
		// clear the screen and initialize stuff
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

		glMatrixMode(GL_MODELVIEW);
		glLoadIdentity();

		glDisable(GL_DEPTH_TEST);
		glDisable(GL_LIGHTING);

		draw_background();


		glEnable(GL_LIGHTING);
		glEnable(GL_DEPTH_TEST);

		glLoadIdentity();
		glTranslated(m_x, m_y, -10.0);
		glRotated(m_angle, -1.0, 0.0, 0.0);

		glBindTexture(GL_TEXTURE_2D, m_wood);
		glCallList(m_hammer_handle);
		glBindTexture(GL_TEXTURE_2D, 0);
		glCallList(m_hammer_topper);
	}

	void KeyDown(SDLKey key, SDLMod mod, int ch)
	{
		// Esc or Alt+F4 quits...
		if ( (key == SDLK_ESCAPE) || ((key == SDLK_F4) && (0 != (mod & KMOD_ALT))) )
		{
			PostQuit();
		}
	}

	void MouseDown(int button, int x, int y) 
	{
		if ( m_state == idle )
		{
			m_state = downswing;
			BeginBusyState();	
		}
	}

	void MouseUp(int button, int x, int y)
	{
	}

	void MouseMove(int button_state, int x, int y, int x_rel, int y_rel)
	{
		m_mousex = (16.0 * x) / m_Width - 8.0;
		m_mousey = (12.0 * (m_Height - y - 1)) / m_Height - 6.0;
		
		m_x = (24.0 * x) / m_Width - 12.0;
		m_y = (18.0 * (m_Height - y - 1)) / m_Height - 9.0 - 8.25;

		Invalidate();
	}

};

// returns power of 2 greater than or equal to num
int p2ge(int num)
{
	int p2 = 1;
	while(p2 < num)
		p2 <<= 1;
	return p2;
}

int main(int argc, char **argv)
{
	int w = 1024, h = 768, bpp = 16, full = 1;
	GLfloat tw = 1.0f, th = 1.0f;

#ifdef WIN32
	w = GetSystemMetrics(SM_CXSCREEN);
	h = GetSystemMetrics(SM_CYSCREEN);
	int W = p2ge(w), H = p2ge(h);

	g_picBackground.init(W, H);
	tw = (GLfloat)w / W;
	th = (GLfloat)h / H;

	// capture the screen
	HDC hScreenDC = GetDC(NULL);
	BITMAPINFO bi;
	ZeroMemory(&bi, sizeof bi);
	bi.bmiHeader.biSize = sizeof bi.bmiHeader;
	bi.bmiHeader.biWidth = w;
	bi.bmiHeader.biHeight = -h;
	bi.bmiHeader.biPlanes = 1;
	bi.bmiHeader.biBitCount = 24;
	bi.bmiHeader.biCompression = BI_RGB;
	bi.bmiHeader.biSizeImage = w * h * 3;
	LPVOID pData;
	HBITMAP hBmp = CreateDIBSection(hScreenDC, &bi, DIB_RGB_COLORS,
		&pData, NULL, 0);
	HDC hMemDC = CreateCompatibleDC(hScreenDC);
	HBITMAP hOldBmp = (HBITMAP)SelectObject(hMemDC, hBmp);
	BitBlt(hMemDC, 0, 0, w, h, hScreenDC, 0, 0, SRCCOPY);

	// copy DIB to PxM, fixing RGB order
	int pixels = w * h;
	unsigned char *src = (unsigned char *)pData;
	unsigned char *row = g_picBackground.getPixels();
	for(int y = 0; y < h; y++)
	{
		unsigned char *dest = row;
		for(int x = 0; x < w; x++)
		{
			dest[0] = src[2];
			dest[1] = src[1];
			dest[2] = src[0];
			src += 3;
			dest += 3;
		}
		row += g_picBackground.getPitch();
	}

	DeleteObject(SelectObject(hMemDC, hOldBmp));
	DeleteDC(hMemDC);
#else
	FILE *f = popen("xwd -silent -root | convert xwd:- ppm:-", "r");
	if ( !f )
	{
		fprintf(stderr, "Screen capture failed.  (Requires xwd from X11 and convert from ImageMagick.)\n");
		return 1;
	}
	PxM screen;
	int res = screen.load_PxM(f);
	pclose(f);
	if ( !res )
	{
		fprintf(stderr, "Failed to receive a valid PPM from xwd | convert\n");
		return 1;
	}

	w = screen.getWidth();
	h = screen.getHeight();
	int W = p2ge(w), H = p2ge(h);

	g_picBackground.init(W, H);
	tw = (GLfloat)w / W;
	th = (GLfloat)h / H;

	unsigned char *src = screen.getPixels();
	unsigned char *dest = g_picBackground.getPixels();
	for(int row = 0; row < h; row++)
	{
		memcpy(dest, src, screen.getPitch());
		src += screen.getPitch();
		dest += g_picBackground.getPitch();
	}
#endif

	banger Banger(tw, th);

	if ( Banger.InitViewport(w, h, bpp, full, "banger") )
	{
#ifdef WIN32
		// On dual monitor systems, the window doesn't necessarily cover exactly
		// one monitor.  Coerce it a little.
		// (We don't use full screen on Win32 because it usually causes a screen blank
		// while starting up, which somewhat ruins the effect)
		SDL_SysWMinfo wmi;
		SDL_VERSION(&wmi.version);
		SDL_GetWMInfo(&wmi);
		MoveWindow(wmi.window, 0, 0, w, h, FALSE);
#endif
		Banger.Initialize();
		Banger.Run();
	}

	return 0;
}
