// ps-gl.cpp
// (c) 2004 by Jeremy Stanley
// This is free software licensed under the GNU GPL.  See license.txt for details.


#include <stdio.h>
#include "ps-gl.h"
#include "PxM.h"	// texture loading (inc. JPEG)

GLuint CreateTexture(PxM &pic, int width /* = -1 */, int height /* = -1 */, bool lum /* = false */)
{
	if (width == -1) width = pic.getWidth();
	if (height == -1) height = pic.getHeight();

	// shrink the texture to fit on the hardware if needed (e.g., 3dfx Doodoo-N cards)
	int maxtex;
	glGetIntegerv(GL_MAX_TEXTURE_SIZE, &maxtex);
	if (width > maxtex || height > maxtex)
	{
		fprintf(stderr, "CreateTexture(): warning: I had to shrink the texture to accomodate your prehistoric video hardware.\n");
	}
	if (width > maxtex)
		width = maxtex;
	if (height > maxtex)
		height = maxtex;

	GLuint texname;
	glGenTextures(1, &texname);
	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
	glPixelStorei(GL_UNPACK_ALIGNMENT, 1);

	int i, j, col[3];
	int Bpp = lum ? 1 : 4;
	GLubyte *buf = new GLubyte[width * height * Bpp], *pTex = buf;

	if (width == pic.getWidth() && height == pic.getHeight())
	{
		// copy the texture

		for(i = height - 1; i >= 0; i--)
		{
			for(j = 0; j < width; j++)
			{
				pic.pixel(j, i, col);

				if (lum)
				{
					*pTex++ = (GLubyte) col[0];
				}
				else
				{
					*pTex++ = (GLubyte) col[0];
					*pTex++ = (GLubyte) col[1];
					*pTex++ = (GLubyte) col[2];
					*pTex++ = (GLubyte) 0xFF;
				}
			}
		}
	}
	else
	{
		// scale and copy

		double u = 0.5, v = (double)pic.getHeight() - 0.5;
		double du = (double)pic.getWidth() / width;
		double dv = -(double)pic.getHeight() / height;

		for (i = 0; i < height; i++)
		{
			u = 0.0;
			v += dv;

			for (j = 0; j < width; j++)
			{
				pic.bilinearpix(u, v, col);

				if (lum)
				{
					*pTex++ = (GLubyte) col[0];
				}
				else
				{
					*pTex++ = (GLubyte) col[0];
					*pTex++ = (GLubyte) col[1];
					*pTex++ = (GLubyte) col[2];
					*pTex++ = (GLubyte) 0xFF;
				}
				u += du;
			}
		}
	}

	glBindTexture(GL_TEXTURE_2D, texname);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	if (lum)
	{
		glTexImage2D(GL_TEXTURE_2D, 0, GL_INTENSITY4, width, height,
			0, GL_LUMINANCE, GL_UNSIGNED_BYTE, buf);
	}
	else
	{
		glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, width, height,
			0, GL_RGBA, GL_UNSIGNED_BYTE, buf);
	}

	// un-bind the texture
	glBindTexture(GL_TEXTURE_2D, 0);
	
	delete[] buf;
	return texname;
}

GLuint CreateTexture(const char *imgfile, int width /*= -1*/, int height /*= -1*/, bool lum /*= false*/)
{
	PxM pic;
	if (!pic.load(imgfile))
	{
		return 0;
	}

	return CreateTexture(pic, width, height, lum);
}

bool IsExtensionSupported(const char *extension)
{
	const char *extensions = (const char *)glGetString(GL_EXTENSIONS);

	const char *ext = strstr(extensions, extension);
	while (ext)
	{
		if ((ext > extensions) && !isspace(ext[-1]))
		{
			break;			
		}

		int len = strlen(extension);
		int tot_len = strlen(extensions);
		if ((ext + len < extensions + tot_len - 1)
			&& !isspace(ext[len]))
		{
			break;
		}
	
		fprintf(stderr, "%s: OK\n", extension);
		return true;
	}

	fprintf(stderr, "%s: unsupported\n", extension);
	return false;
}
