// ps-sdl_gl_app.h
// (c) 2004 by Jeremy Stanley
// This is free software licensed under the GNU GPL.  See license.txt for details.

#ifndef PS_SDLGLAPP_H
#define PS_SDLGLAPP_H

#include "ps-gl.h"
#include "SDL.h"
#include <string>

class psSdlGlApp
{
public:
	psSdlGlApp();
	virtual ~psSdlGlApp();

	bool InitViewport(int width, int height, int depth, int mode, const char *caption);
	void Run();

	void Invalidate();

protected:
	// overridables
	virtual void KeyDown(SDLKey key, SDLMod mod, int ch) {}
	virtual void KeyUp(SDLKey key, SDLMod mod) {}
	virtual void Reshape(int w, int h) {}
	virtual void UpdateBusyState(double frametime) {}
	virtual void Draw() {}
	virtual void MouseDown(int button, int x, int y) {}
	virtual void MouseUp(int button, int x, int y) {}
	virtual void MouseMove(int button_state, int x, int y, int x_rel, int y_rel) {}
	virtual bool OnQuit() { return true; }	/* return false in your implementation to prevent quitting */
	virtual void OnUserEvent(int code, void *data1, void *data2) {}
	virtual void OnCommand(int dialog_id, int widget_id, void *data) {}

	// state members
	GLuint m_Width, m_Height;	// width and height of viewport
	int m_mode;	// 0 = windowed; 1 = full screen; 2 = game mode
	std::string m_caption;		// window caption

protected:
	void BeginBusyState();
	void EndBusyState();
	void ResetBusyState();

	void PostQuit();

private:
	void do_display(void);
	int  ProcessEvents();
	void DispatchEvent(SDL_Event &event);

	// speed calibration stuff
	int m_busy;
	int m_timestamps;
	bool m_quit_req;
	double m_Frametime;
	bool m_need_redraw;
	void calibrate_speed();
};

extern psSdlGlApp *theApp;

#endif // PS_SDLGLAPP_H

